#include <stdio.h>   /* Standard input/output definitions */
#include <string.h>  /* String function definitions */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <getopt.h>  /* Getopt definitions */
#include <stdlib.h>  /* Standard library definitions */
#include <assert.h>  /* Assert library definitions */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/select.h>
#include <time.h>

// Variables avaliable to T-series
// according to nad_commands_TSeries_1.10.pdf
#define VAR_DEVICE_ID     "\x14"
#define VAR_POWER         "\x15"
#define VAR_VOLUME        "\x17"
#define VAR_MUTE          "\x18"
#define VAR_ZONE_VOLUME   "\x1B"
#define VAR_ZONE_MUTE     "\x1C"
#define VAR_TAPE_MONITOR  "\x1F"
#define VAR_TREBLE        "\x20"
#define VAR_BASS          "\x21"
#define VAR_ASIGNAL_TYPE  "\x27"
#define VAR_SURROUND_MODE "\x28"
#define VAR_VINPUT_TYPE   "\x29"
#define VAR_TUNER_BAND    "\x2B"
#define VAR_AM_FREQ       "\x2C"
#define VAR_FM_FREQ       "\x2D"
#define VAR_FM_MUTE       "\x2F"
#define VAR_ZONE_POWER    "\x34"
#define VAR_CURRENT_INPUT "\x35"
#define VAR_ZONE_INPUT    "\x36"
#define VAR_INPUT_NAME    "\x37"
#define VAR_AM_PRESET     "\x38"
#define VAR_FM_PRESET     "\x39"
#define VAR_SPEAKER_A     "\x3A"
#define VAR_SPEAKER_B     "\x3B"
#define VAR_TONE_DEFEAT   "\x3C"
#define VAR_STEREO_MODE   "\x40"
#define VAR_CROSSOVER     "\x41"

// request commands
#define CMD_REQ_GET "\x14"
#define CMD_REQ_SET "\x15"
#define CMD_REQ_REMOTE "\x16"
#define CMD_REQ_SW_VERSION "\x17"
#define CMD_REQ_STORE_AV_PRESET "\x19"

// response commands
#define CMD_RESP_VAL "\x14"
#define CMD_RESP_SW_VERSION "\x17"
#define CMD_RESP_AV_PRESET_STORED "\x19"
#define CMD_RESP_RDS_PS "\x1B"
#define CMD_RESP_RDS_RT "\x1C"

// remote control signals
#define REMOTE_AV_PRESET_1 "\x3A"
#define REMOTE_AV_PRESET_2 "\x3B"
#define REMOTE_AV_PRESET_3 "\xC3"
#define REMOTE_AV_PRESET_4 "\xC4"
#define REMOTE_AV_PRESET_5 "\xC6"
#define REMOTE_AM "\x82"
#define REMOTE_CD "\x85"
#define REMOTE_CENTER_DOWN "\xD5"
#define REMOTE_CENTER_UP "\xD0"
#define REMOTE_CURSOR_DOWN "\x28"
#define REMOTE_CURSOR_ENTER "\x32"
#define REMOTE_CURSOR_LEFT "\x29"
#define REMOTE_CURSOR_RIGHT "\x2A"
#define REMOTE_CURSOR_UP "\x27"
#define REMOTE_DISC "\x89"
#define REMOTE_DISPLAY "\x26"
#define REMOTE_EXTERNAL_7_1 "\x2E"
#define REMOTE_DISPLAY "\x26"
#define REMOTE_FM "\x81"
#define REMOTE_LATE_NIGHT "\x2F"
#define REMOTE_LEVELS "\x2B"
#define REMOTE_MUTE "\x94"
#define REMOTE_POWER_OFF "\xC8"
#define REMOTE_POWER_ON "\x25"
#define REMOTE_POWER_TOGGLE "\x80"
#define REMOTE_PRESET_DOWN "\xD1"
#define REMOTE_PRESET_UP "\xD2"
#define REMOTE_SEARCH_DOWN "\x86"
#define REMOTE_SEARCH_UP "\x87"
#define REMOTE_SLEEP "\x20"
#define REMOTE_SOURCE_DOWN "\x1D"
#define REMOTE_SOURCE_UP "\x1A"
#define REMOTE_SPEAKER_A "\xCE"
#define REMOTE_SPEAKER_B "\xCF"
#define REMOTE_SUB_WOOFER_DOWN "\xD9"
#define REMOTE_SUB_WOOFER_UP "\xD8"
#define REMOTE_SURROUND_DOWN "\xD7"
#define REMOTE_SURROUND_MODE "\xCC"
#define REMOTE_SURROUND_UP "\xD6"
#define REMOTE_TAPE "\x8D"
#define REMOTE_TEST "\xCD"
#define REMOTE_TONE "\x39"
#define REMOTE_TONE_DEFEAT "\xDC"
#define REMOTE_TUNE_DOWN "\xD3"
#define REMOTE_TUNE_UP "\xD4"
#define REMOTE_TUNER "\xDD"
#define REMOTE_TUNER_0 "\xC7"
#define REMOTE_TUNER_1 "\x8A"
#define REMOTE_TUNER_2 "\x8E"
#define REMOTE_TUNER_3 "\x92"
#define REMOTE_TUNER_4 "\x96"
#define REMOTE_TUNER_5 "\x8B"
#define REMOTE_TUNER_6 "\x8F"
#define REMOTE_TUNER_7 "\x93"
#define REMOTE_TUNER_8 "\x97"
#define REMOTE_TUNER_9 "\x98"
#define REMOTE_TUNER_ENTER "\xC5"
#define REMOTE_VIDEO_1_DVD "\xC2"
#define REMOTE_VIDEO_2_SAT "\xC0"
#define REMOTE_VIDEO_3_VCR "\xC1"
#define REMOTE_VOLUME_DOWN "\x8C"
#define REMOTE_VOLUME_UP "\x88"

// argument list for response handlers
#define HANDLER_ARGS fdinfo * fdi, char * data

void model_message(HANDLER_ARGS);
void vol_message(HANDLER_ARGS);
void speaker_message(HANDLER_ARGS);
void get_time_message(HANDLER_ARGS);
void set_time_message(HANDLER_ARGS);

// define handlers
extern struct message_handler {
  char * trigger;
  void (*handler) (HANDLER_ARGS);
} handlers[];


