#include <stdio.h>   /* Standard input/output definitions */
#include <string.h>  /* String function definitions */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <getopt.h>  /* Getopt definitions */
#include <stdlib.h>  /* Standard library definitions */
#include <assert.h>  /* Assert library definitions */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/select.h>
#include <time.h>

#define DEFAULT_PORT "/dev/ttyS0"
#define DEFAULT_LIMIT -20
#define BAUDRATE B9600

#define UNCH unsigned char

// request cmds
#define CMD_REQ_MODEL      "\x02\x14\x14"
#define CMD_REQ_VOL        "\x02\x14\x17"
#define CMD_REQ_VOL_UP     "\x02\x16\x88"
#define CMD_REQ_VOL_DOWN   "\x02\x16\x8C"
#define CMD_REQ_ALL        "\x02\x14\xFF"

// response cmds
#define CMD_RESP_VOL_CHANGE "\x02\x14\x17"
#define CMD_RESP_MODEL_INFO "\x02\x14\x14"
#define CMD_RESP_SPEAKER_A  "\x02\x14\x3A"
#define CMD_RESP_SPEAKER_B  "\x02\x14\x3B"

// argument list for response handlers
#define HANDLER_ARGS char * data

// misc macros
#define AZ(foo) do { assert((foo) == 0); } while (0)
#define Vprintf(...) if ( verbose_flag ) { printf(__VA_ARGS__); }
#define Vprintf1(...) if ( verbose_flag >= 1 ) { printf(__VA_ARGS__); }
#define Vprintf2(...) if ( verbose_flag >= 2 ) { printf(__VA_ARGS__); }
#define Vprintf3(...) if ( verbose_flag >= 3 ) { printf(__VA_ARGS__); }

/* Serial port */
char * port;

/* Flag set by `--verbose'. */
UNCH verbose_flag;

/* Used to store `--limit'. */
signed char limit;
signed char speakers[2];

struct message_handler {
  char * trigger;
  void (*handler) (HANDLER_ARGS);
};

struct config_entry {
  UNCH hour;
  UNCH wday;
  signed char limit;
  UNCH speaker[2];
};



int fd; /* File descriptor for the port */

/*
 * 'open_port()' - Open serial port 1.
 *
 * Returns the file descriptor on success or -1 on error.
 */
int open_port(void)
{
  fd = open(port, O_RDWR | O_NOCTTY | O_NDELAY);
  if (fd == -1)
  {
    /*
     * Could not open the port.
     */

    UNCH buf[60];
    snprintf(buf, 60, "open_port: Unable to open port %s", port);
    perror(buf);
    abort();
  }

  // set nonblocking
  fcntl(fd, F_SETFL, FNDELAY);

  // set blocking
  //fcntl(fd, F_SETFL, 0);

  // get current options
  struct termios options;
  tcgetattr(fd, &options);

  // No parity (8N1)
  options.c_cflag &= ~PARENB;
  options.c_cflag &= ~CSTOPB;
  options.c_cflag &= ~CSIZE;
  options.c_cflag |= CS8;
  
  // raw input mode
  options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

  // raw output mode
  options.c_oflag &= ~OPOST;

  // disable hardware flow control
  options.c_cflag &= ~CRTSCTS;

  // disable software flow control
  options.c_iflag &= ~(IXON | IXOFF | IXANY);

  // set baud
  cfsetispeed(&options, BAUDRATE);
  cfsetospeed(&options, BAUDRATE);

  // set options
  tcsetattr(fd, TCSANOW, &options);

  return (fd);
}

void close_port(int fd)
{
  AZ(close(fd));
}

void eusage(char * name, int exit_flag)
{
  printf("Usage: %s [OPTIONS]... [PORT]\n", name);
  if (exit_flag) exit(1);
}

void usage(char * name)
{
  eusage(name, 1);
}

void help(char * name)
{
  eusage(name, 0);
  printf("\n");
  printf("Mandatory arguments to long options are mandatory for short options too.\n");
  printf("  -h, --help                 shows this help\n");
  printf("  -c, --config               sets configuration file\n");
  printf("  -v, --verbose              enables verbose output\n");
  printf("\n");
  exit(0);
}

int parse_options(int argc, char * const argv[])
{
  static struct option long_options[] = 
    {
      {"verbose", no_argument,       0, 'v'},
      {"help",    no_argument,       0, 'h'},
      {"limit",   required_argument, 0, 'l'},
      {0, 0, 0, 0}
    };

  int c;

  verbose_flag = 0;
  while (1) {
    int option_index = 0;

    c = getopt_long(argc, argv, "vhl:", long_options, &option_index);

    if (c == -1)
      break;

    switch (c) {
      case 'v':
        verbose_flag++;
        break;
      case 'h':
        help(argv[0]); 
        break;
      case 'l':
        limit = atoi(optarg);
        break;
      case '?':
        usage(argv[0]);
        break;
      default:  
        abort();
    }
  }

  if (optind + 1 == argc) {
    port = argv[optind++];
  }

  /* Print any remaining command line arguments (not options). */
  if (optind < argc) {
    printf ("non-option ARGV-elements: ");
    while (optind < argc)
      printf ("%s ", argv[optind++]);
    putchar ('\n');
  }

  /* Instead of reporting `--verbose'
   * and `--brief' as they are encountered,
   * we report the final status resulting from them. */
  Vprintf1("Verbose mode enabled at level %d\n", verbose_flag)
  Vprintf1("limit = %d\n", limit)
  Vprintf1("port = \"%s\"\n", port)

}

char chksum(char * data)
{
  int i;
  UNCH foo = '\x00';
  Vprintf3("chksum:");
  for (i=1;i<=data[0];i++) {
    Vprintf3(" %02X", (UNCH) data[i]);
    foo += data[i];
  }
  foo = 0xff - foo;
  Vprintf3(" = %02X\n", foo);
  return foo;
}

void sendcmd(char * data)
{
  UNCH sum = chksum(data);
  UNCH buffer[15];
  int i = 0;
  int j = 0;
  buffer[i++] = '\x01';
  for (j = 1; j <= data[0]; j++) {
    if ((UNCH) data[j] < '\20') {
      buffer[i++] = '\x5E';
      buffer[i++] = data[j] ^ '\x40';
    } else if (data[j] == '\x5E') {
      buffer[i++] = '\x5E';
      buffer[i++] = '\x5E';
    } else {
      buffer[i++] = data[j];
    }
  }
  buffer[i++] = '\x02';
  buffer[i++] = sum;
  
  Vprintf("Sending:");
  for (j=0;j<i;j++) Vprintf(" %02X",buffer[j]);
  Vprintf("\n");

  ssize_t ret = write(fd, buffer, i);
  if (ret == -1) {
    perror("write");
  }

  //return "\x14\x14\x54\x37\x36\x33";
}

void speaker_helper(UNCH set, UNCH speaker, UNCH flag) {
  UNCH buffer[5];
  bzero(buffer, 5);
  buffer[0] = set ? 3 : 2;
  buffer[1] = 20 + (UNCH) set;
  buffer[2] = 58 + (UNCH) speaker;
  buffer[3] = (UNCH) flag;
  sendcmd(buffer); 
}

void set_speaker_a(UNCH flag) {
  speaker_helper(1, 0, flag);
}

void set_speaker_b(UNCH flag) {
  speaker_helper(1, 1, flag);
}

void req_speaker_a() {
  speaker_helper(0, 0, 0);
}

void req_speaker_b() {
  speaker_helper(0, 1, 0);
}

void set_volume(int vol) {
  UNCH buffer[5];
  bzero(buffer, 5);
  buffer[0] = 3;
  buffer[1] = 21;
  buffer[2] = 23;
  buffer[3] = (signed char) vol;
  sendcmd(buffer); 
}

// MESSAGE HANDLERS

void model_message(HANDLER_ARGS)
{
  int i;
  printf("Connected to \"");
  for (i = 3; i <= data[0]; i++) {
    printf("%c", data[i]);
  }
  printf("\"\n");
  sendcmd(CMD_REQ_VOL);
  req_speaker_a();
  req_speaker_b();
}

void vol_message(HANDLER_ARGS)
{
  signed char vol = data[3];
  printf("volume = %d dB\n", vol);

  if (vol > limit) {
    printf("volume (%d dB) over limit (%d dB)\n", vol, limit);
    printf("enforcing limit\n");
    set_volume(limit);
  }
}

void speaker_message(HANDLER_ARGS)
{
  UNCH speaker = data[2] - '\x3A';
  UNCH value = data[3];
  printf("speaker %X = %d\n", ((UNCH) speaker) + 10, value);

  if (value == 1 && speakers[speaker] == 0) {
    printf("speaker %X is ON but should be OFF\n", ((UNCH) speaker) + 10);
    printf("enforcing limit\n");
    speaker_helper(1, (UNCH) speaker, 0);
  }
}

int readcmd(UNCH * cmd)
{
  UNCH buffer[15];
  unsigned short int buffer_pos = 0;
  UNCH cbuffer[20];
  ssize_t cbytes;
  UNCH record = 0;
  bzero(buffer, 15);

  while (1) {
    cbytes = read(fd, cbuffer, 1);
    if (cbytes == -1) {
      if (errno == EAGAIN) {
        continue;
      }
      perror("read");
      sleep(1);
      continue;
    }
    //printf("%d, %02X\n", cbytes, (UNCH) cbuffer[0]);
    switch (cbuffer[0]) {
      case '\x01':
        record = 1;
        break;
      case '\x02':
        if (record == 1) {
          record = 2;
        }
        break;
      default:
        if (record == 1) {
          buffer[buffer_pos++] = cbuffer[0];
        } else if (record == 2) {
          Vprintf("Input:");
          int i; for (i=0;i<strlen(buffer);i++) Vprintf(" %02X",buffer[i]);
          Vprintf("\n");
          
          // decode;
          bzero(cmd, 15);
          int j = 1;
          for (i = 0; i < strlen(buffer); i++) {
            if (buffer[i] == '\x5E' && i+1 < strlen(buffer)) {
              i++;
              if (buffer[i] == '\x5E') {
                cmd[j++] = buffer[i];
              } else {
                cmd[j++] = buffer[i] & 191;
              }
            } else {
              cmd[j++] = buffer[i];
            }
          }
          cmd[0] = j - 1;
          
          Vprintf("CMD:");
          for (i=1;i<=cmd[0];i++) Vprintf(" %02X", cmd[i]);
          Vprintf(" = %X\n", cbuffer[0]);

          if ((UNCH) chksum(cmd) == (UNCH) cbuffer[0]) {
            return strlen(cmd);
          } else {
            Vprintf("chksum: %02X != %02X\n", (UNCH) chksum(cmd), (UNCH) cbuffer[0]);
            return -1;
          }
        }
        break;
    }
  }
}

void processinput() {
  // define message handlers
  static struct message_handler handlers[] =
    {
      {CMD_RESP_VOL_CHANGE, vol_message},
      {CMD_RESP_MODEL_INFO, model_message},
      {CMD_RESP_SPEAKER_A,  speaker_message},
      {CMD_RESP_SPEAKER_B,  speaker_message},
      {0, 0}
    };

  UNCH response[15];
  int i, j;
  int found = 0;
  readcmd(response);
  Vprintf("Looking for handler to:");
  int k; for (k=1;k<=response[0];k++) Vprintf(" %02X",response[k]);
  Vprintf("\n");
  for (i = 0; handlers[i].trigger && !found; i++) {
    if (response[0] < handlers[i].trigger[0])
      continue;
    for (j = 1; j <= handlers[i].trigger[0] + 1; j++) {
      if (handlers[i].trigger[0] == j-1) {
        //Vprintf("Calling handler for:");
        //int l; for (l=1;l<j;l++) Vprintf(" %02X",response[l]);
        //Vprintf("\n");
        handlers[i].handler(response);
        found = 1;
        break;
      } else if (handlers[i].trigger[j] != response[j]) {
        break;
      }
    }
  }
}

void change_limit(int nl) {
  if (limit != nl) {
    limit = nl;
    sendcmd(CMD_REQ_VOL);
  }
}

void change_limit_speakers(UNCH i, UNCH v) {
  if (speakers[i] != v) {
    speakers[i] = v;
    speaker_helper(0, i, 0);
  }
}

void adjustlimit() {
  static struct config_entry config[] = {
    {1,  8,   5,  {1, 1}},
    {1, 22, -20,  {0, 1}},
    {2,  8,   5,  {1, 1}},
    {2, 22, -20,  {0, 1}},
    {3,  8,   5,  {1, 1}},
    {3, 22, -20,  {0, 1}},
    {4,  8,   5,  {1, 1}},
    {4, 22, -20,  {0, 1}},
    {5,  8,   5,  {1, 1}},
    {6,  2, -20,  {0, 1}},
    {7,  2, -20,  {0, 1}},
    {7, 22, -20,  {0, 1}},
    {0,  0,   0,  {0, 0}}
  };
  static UNCH config_cur = 0;
 
  time_t s =  time(NULL);
  struct tm curtm;
  localtime_r(&s, &curtm);
 
  UNCH max_config = 0;
  unsigned short int max_hours = 0;

  UNCH i;
  for (i = 0; config[i].wday; i++) {
    signed short int diff_hours;
    diff_hours = config[i].hour - curtm.tm_hour - 1;
    diff_hours += 24 * ((config[i].wday == 7 ? 0 : config[i].wday) - curtm.tm_wday);

    if (diff_hours < 0) {
      diff_hours += 7 * 24;
    }

    if (diff_hours > max_hours) {
      max_hours = diff_hours;
      max_config = i;
    }
  }

  if (config_cur != max_config) {
    config_cur = max_config;
    change_limit(config[config_cur].limit);
    change_limit_speakers(0, config[config_cur].speaker[0]);
    change_limit_speakers(1, config[config_cur].speaker[1]);
  }
}

void idleloop()
{
  Vprintf("Initiating idle loop\n");

  /* Do the select */
  while (1) {
    /* Initialize the input set */
    fd_set input;
    FD_ZERO(&input);
    FD_SET(fd, &input);

    //int max_fd = (sock > fd ? sock : fd) + 1;
    int max_fd = fd + 1;

    /* Initialize the timeout structure */
    struct timeval timeout;
    timeout.tv_sec  = 10;
    timeout.tv_usec = 0;

    // call select(2)
    int n = select(max_fd, &input, NULL, NULL, &timeout);

    /* See if there was an error */
    if (n < 0) {
      perror("select failed");
    } else if (n == 0) {
      Vprintf("select(2): TIMEOUT\n");
    } else {
      /* We have input */
      if (FD_ISSET(fd, &input)) {
        processinput();
      }
      //if (FD_ISSET(sock, &input))
      //process_socket();
    }
    
    adjustlimit();
  }
}

int main(int argc, char * const argv[])
{
  // handle config
  limit = DEFAULT_LIMIT;
  port = DEFAULT_PORT;
  speakers[0] = 1;
  speakers[1] = 1;
  parse_options(argc, argv);

  // open the port
  open_port();

  // request model info
  sendcmd(CMD_REQ_MODEL);
  //sendcmd(CMD_REQ_ALL); // ALL includes model

  // initiate idle loop
  idleloop();

  close_port(fd);

  Vprintf("Exiting...\n");
}
