/* 
 * Orginal networking code:
 * Non blocking server 
 * By Vic Metcalfe (vic@acm.org)
 * For the unix-socket-faq
 */


#include "main.h"
#include "socket.h"
#include "verbose.h"
#include "rs232.h"

int sock;                       /* The socket file descriptor for our "listening" socket */
fdinfo connectlist[MAX_CONNS]; /* Array of connected sockets so we know who we are talking to */
fd_set socks;                 /* Socket file descriptors we want to wake up for, using select() */
int highsock;                /* Highest #'d file descriptor, needed for select() */

void setnonblocking(int sock)
{
  int opts;

  opts = fcntl(sock,F_GETFL);
  if (opts < 0) {
    perror("fcntl(F_GETFL)");
    exit(EXIT_FAILURE);
  }
  opts = (opts | O_NONBLOCK);
  if (fcntl(sock,F_SETFL,opts) < 0) {
    perror("fcntl(F_SETFL)");
    exit(EXIT_FAILURE);
  }
  return;
}

/* This is just like the write() system call, accept that it will
*    make sure that all data is transmitted. */
int sock_write(sockfd, buf, count)
  int sockfd;
  const char *buf;
  size_t count;
{
  size_t bytes_sent = 0;
  int this_write;

  while (bytes_sent < count) {
    do this_write = write(sockfd, buf, count - bytes_sent);
    while ( (this_write < 0) && (errno == EINTR) );
      if (this_write <= 0)
        return this_write;
    bytes_sent += this_write;
    buf += this_write;
  }
  return count;
}

void socket_build_select_list(fd_set * socks, int * highsock)
{
  FD_SET(sock,socks);

  if (sock > *highsock) {
    *highsock = sock;
  }
  int listnum;
  for (listnum = 0; listnum < MAX_CONNS; listnum++) {
    if (connectlist[listnum].fd != 0) {
      FD_SET(connectlist[listnum].fd, socks);
      if (connectlist[listnum].fd > *highsock) {
        *highsock = connectlist[listnum].fd;
      }
    }
  }
}

void socket_handle_new_connection()
{
  int listnum;       /* Current item in connectlist for for loops */
  int connection; /* Socket file descriptor for incoming connections */

  /* We have a new connection coming in!  We'll
  *  try to find a spot for it in connectlist. */
  connection = accept(sock, NULL, NULL);
  if (connection < 0) {
    perror("accept");
    exit(EXIT_FAILURE);
  }
  setnonblocking(connection);
  for (listnum = 0; (listnum < MAX_CONNS) && (connection != -1); listnum++) {
    if (connectlist[listnum].fd == 0) {
      printf("\nConnection accepted:   FD=%d; Slot=%d\n", connection, listnum);
      connectlist[listnum].fd = connection;
      connectlist[listnum].type = FD_TYPE_CONN;
      connection = -1;
    }
  }
  if (connection != -1) {
    /* No room left in the queue! */
    printf("\nNo room left for new client.\n");
    // TODO replace with appropriate response
    write(connection, "Sorry, this server is too busy. Try again later!\r\n", strlen("Sorry, this server is too busy. Try again later!\r\n"));
    AZ(shutdown(connection, SHUT_RDWR));
    AZ(close(connection));
  }
}

void close_socket(int listnum)
{
  if (connectlist[listnum].fd == 0) return;
  Vprintf3("Terminating connection (FD: %d)\n", connectlist[listnum].fd);
  shutdown(connectlist[listnum].fd, SHUT_RDWR);
  AZ(close(connectlist[listnum].fd));
  connectlist[listnum].fd = 0;
}

void close_sockets()
{
  int listnum;
  for (listnum = 0; listnum < MAX_CONNS; listnum++) {
    close_socket(listnum);
  }
}

void close_socket_by_fdi(fdinfo * fdi)
{
  int listnum;
  for (listnum = 0; listnum < MAX_CONNS; listnum++) {
    if (connectlist[listnum].fd == fdi->fd) {
      close_socket(listnum);
    }
  }
}

void socket_broadcast(char * data)
{
  int listnum;
  for (listnum = 0; listnum < MAX_CONNS; listnum++) {
    if (connectlist[listnum].fd > 0) {
      Vprintf3("Sending it back\n");
      sendcmd(&connectlist[listnum], data);
    }
  }
}

/*
void deal_with_data(int listnum)
{
  char buffer[80];     //Buffer for socket reads
  char *cur_char;      // Used in processing buffer 

  if (sock_gets(connectlist[listnum],buffer,80) < 0) {
    // 
    // Connection closed, close this end
    // and free up entry in connectlist
    //
    printf("\nConnection lost: FD=%d;  Slot=%d\n",
    connectlist[listnum],listnum);
    AZ(close(connectlist[listnum]));
    connectlist[listnum] = 0;
  } else {
    //
    // We got some data, so upper case it
    //       and send it back.
    //
    printf("\nReceived: %s; ",buffer);
    cur_char = buffer;
    while (cur_char[0] != 0) {
      cur_char[0] = toupper(cur_char[0]);
      cur_char++;
    }
    //sock_puts(connectlist[listnum],buffer);
    //sock_puts(connectlist[listnum],"\n");
    printf("responded: %s\n",buffer);
  }
}
*/

void socket_read_selected(fd_set * socks)
{
  int listnum;       /* Current item in connectlist for for loops */

  /* 
   * OK, now socks will be set with whatever socket(s)
   * are ready for reading.  Lets first check our
   * "listening" socket, and then check the sockets
   * in connectlist.
   */

  /*
   * If a client is trying to connect() to our listening
   * socket, select() will consider that as the socket
   * being 'readable'. Thus, if the listening socket is
   * part of the fd_set, we need to accept a new connection.
   */

  if (FD_ISSET(sock, socks))
    socket_handle_new_connection();

  /* Now check connectlist for available data */

  /*
   * Run through our sockets and check to see if anything
   * happened with them, if so 'service' them.
   */

  for (listnum = 0; listnum < MAX_CONNS; listnum++) {
    if (connectlist[listnum].fd != 0) {
      if (FD_ISSET(connectlist[listnum].fd, socks)) {
        Vprintf3("FD %d: have been selected!\n", connectlist[listnum].fd);
        processinput(&connectlist[listnum]);
      }
    }
  }
}

int socket_open() 
{
  struct sockaddr_in server_address; /* bind info structure */
  int reuse_addr = 1;  /* Used so we can re-bind to our port while a previous connection is still in TIME_WAIT state. */

  /* Obtain a file descriptor for our "listening" socket */
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    perror("socket");
    exit(EXIT_FAILURE);
  }
  /* So that we can re-bind to it without TIME_WAIT problems */
  setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuse_addr,
  sizeof(reuse_addr));

  /* Set socket to non-blocking with our setnonblocking routine */
  setnonblocking(sock);

  /* Get the address information, and bind it to the socket */
  memset((char *) &server_address, 0, sizeof(server_address));
  server_address.sin_family = AF_INET;
  server_address.sin_addr.s_addr = htonl(INADDR_ANY);
  server_address.sin_port = htons(BIND_PORT);
  if (bind(sock, (struct sockaddr *) &server_address, sizeof(server_address)) < 0 ) {
    perror("bind");
    close(sock);
    exit(EXIT_FAILURE);
  }

  /* Set up queue for incoming connections. */
  listen(sock, LISTEN_QUEUE);
  memset((char *) &connectlist, 0, sizeof(connectlist));
  return 1;
}

/*
{
   
   
    build_select_list();
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    readsocks = select(highsock+1, &socks, (fd_set *) 0, 
    (fd_set *) 0, &timeout);

    if (readsocks < 0) {
      perror("select");
      exit(EXIT_FAILURE);
    }
    if (readsocks == 0) {
      
       * Nothing ready to read, just show that
       * we're alive
       
      printf(".");
      fflush(stdout);
    } else {
      read_socks();
    }
  }
}*/

